@@ -1,6 +1,6 @@
 # vim: ft=python fileencoding=utf-8 sts=4 sw=4 et:
 
-# Copyright 2014-2015 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
+# Copyright 2014-2017 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
 #
 # This file is part of qutebrowser.
 #
@@ -17,4 +17,17 @@
 # You should have received a copy of the GNU General Public License
 # along with qutebrowser.  If not, see <http://www.gnu.org/licenses/>.
 
-"""Utilities and classes regarding to commands."""
+"""In qutebrowser, all keybindings are mapped to commands.
+
+Some commands are hidden, which means they don't show up in the command
+completion when pressing `:`, as they're typically not useful to run by hand.
+
+For command arguments, there are also some variables you can use:
+
+- `{url}` expands to the URL of the current page
+- `{url:pretty}` expands to the URL in decoded format
+- `{clipboard}` expands to the clipboard contents
+- `{primary}` expands to the primary selection contents
+
+It is possible to run or bind multiple commands by separating them with `;;`.
+"""
