@@ -1,6 +1,6 @@
 # vim: ft=python fileencoding=utf-8 sts=4 sw=4 et:
 
-# Copyright 2014-2015 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
+# Copyright 2014-2017 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
 #
 # This file is part of qutebrowser.
 #
@@ -21,11 +21,33 @@
 
 import inspect
 import collections
+import traceback
+import typing
 
-from PyQt5.QtWebKit import QWebSettings
+import attr
 
 from qutebrowser.commands import cmdexc, argparser
-from qutebrowser.utils import log, utils, message, debug, docutils, objreg
+from qutebrowser.utils import log, message, docutils, objreg, usertypes
+from qutebrowser.utils import debug as debug_utils
+from qutebrowser.misc import objects
+
+
+@attr.s
+class ArgInfo:
+
+    """Information about an argument."""
+
+    win_id = attr.ib(False)
+    count = attr.ib(False)
+    hide = attr.ib(False)
+    metavar = attr.ib(None)
+    flag = attr.ib(None)
+    completion = attr.ib(None)
+    choices = attr.ib(None)
+
+    def __attrs_post_init__(self):
+        if self.win_id and self.count:
+            raise TypeError("Argument marked as both count/win_id!")
 
 
 class Command:
@@ -36,49 +58,57 @@ class Command:
         name: The main name of the command.
         maxsplit: The maximum amount of splits to do for the commandline, or
                   None.
-        hide: Whether to hide the arguments or not.
+        deprecated: False, or a string to describe why a command is deprecated.
         desc: The description of the command.
         handler: The handler function to call.
-        completion: Completions to use for arguments, as a list of strings.
         debug: Whether this is a debugging command (only shown with --debug).
         parser: The ArgumentParser to use to parse this command.
-        special_params: A dict with the names of the special parameters as
-                        values.
         flags_with_args: A list of flags which take an argument.
-        _type_conv: A mapping of conversion functions for arguments.
-        _name_conv: A mapping of argument names to parameter names.
-        _needs_js: Whether the command needs javascript enabled
-        _modes: The modes the command can be executed in.
-        _not_modes: The modes the command can not be executed in.
+        no_cmd_split: If true, ';;' to split sub-commands is ignored.
+        backend: Which backend the command works with (or None if it works with
+                 both)
+        no_replace_variables: Don't replace variables like {url}
+        modes: The modes the command can be executed in.
+        _qute_args: The saved data from @cmdutils.argument
         _count: The count set for the command.
         _instance: The object to bind 'self' to.
         _scope: The scope to get _instance for in the object registry.
-
-    Class attributes:
-        AnnotationInfo: Named tuple for info from an annotation.
     """
 
-    AnnotationInfo = collections.namedtuple('AnnotationInfo',
-                                            ['kwargs', 'type', 'name', 'flag',
-                                             'special'])
+    def __init__(self, *, handler, name, instance=None, maxsplit=None,
+                 modes=None, not_modes=None, debug=False, deprecated=False,
+                 no_cmd_split=False, star_args_optional=False, scope='global',
+                 backend=None, no_replace_variables=False):
+        if modes is not None and not_modes is not None:
+            raise ValueError("Only modes or not_modes can be given!")
+        if modes is not None:
+            for m in modes:
+                if not isinstance(m, usertypes.KeyMode):
+                    raise TypeError("Mode {} is no KeyMode member!".format(m))
+            self.modes = set(modes)
+        elif not_modes is not None:
+            for m in not_modes:
+                if not isinstance(m, usertypes.KeyMode):
+                    raise TypeError("Mode {} is no KeyMode member!".format(m))
+            self.modes = set(usertypes.KeyMode).difference(not_modes)
+        else:
+            self.modes = set(usertypes.KeyMode)
+        if scope != 'global' and instance is None:
+            raise ValueError("Setting scope without setting instance makes "
+                             "no sense!")
 
-    def __init__(self, name, maxsplit, hide, instance, completion, modes,
-                 not_modes, needs_js, is_debug, ignore_args,
-                 handler, scope):
-        # I really don't know how to solve this in a better way, I tried.
-        # pylint: disable=too-many-arguments,too-many-locals
         self.name = name
         self.maxsplit = maxsplit
-        self.hide = hide
+        self.deprecated = deprecated
         self._instance = instance
-        self.completion = completion
-        self._modes = modes
-        self._not_modes = not_modes
         self._scope = scope
-        self._needs_js = needs_js
-        self.debug = is_debug
-        self.ignore_args = ignore_args
+        self._star_args_optional = star_args_optional
+        self.debug = debug
         self.handler = handler
+        self.no_cmd_split = no_cmd_split
+        self.backend = backend
+        self.no_replace_variables = no_replace_variables
+
         self.docparser = docutils.DocstringParser(handler)
         self.parser = argparser.ArgumentParser(
             name, description=self.docparser.short_desc,
@@ -91,11 +121,14 @@ class Command:
         self.namespace = None
         self._count = None
         self.pos_args = []
-        self.special_params = {'count': None, 'win_id': None}
         self.desc = None
         self.flags_with_args = []
-        self._type_conv = {}
-        self._name_conv = {}
+
+        # This is checked by future @cmdutils.argument calls so they fail
+        # (as they'd be silently ignored otherwise)
+        self._qute_args = getattr(self.handler, 'qute_args', {})
+        self.handler.qute_args = None
+
         self._inspect_func()
 
     def _check_prerequisites(self, win_id):
@@ -106,21 +139,16 @@ class Command:
         """
         mode_manager = objreg.get('mode-manager', scope='window',
                                   window=win_id)
-        curmode = mode_manager.mode
-        if self._modes is not None and curmode not in self._modes:
-            mode_names = '/'.join(mode.name for mode in self._modes)
+        self.validate_mode(mode_manager.mode)
+
+        if self.backend is not None and objects.backend != self.backend:
             raise cmdexc.PrerequisitesError(
-                "{}: This command is only allowed in {} mode.".format(
-                    self.name, mode_names))
-        elif self._not_modes is not None and curmode in self._not_modes:
-            mode_names = '/'.join(mode.name for mode in self._not_modes)
-            raise cmdexc.PrerequisitesError(
-                "{}: This command is not allowed in {} mode.".format(
-                    self.name, mode_names))
-        if self._needs_js and not QWebSettings.globalSettings().testAttribute(
-                QWebSettings.JavascriptEnabled):
-            raise cmdexc.PrerequisitesError(
-                "{}: This command needs javascript enabled.".format(self.name))
+                "{}: Only available with {} "
+                "backend.".format(self.name, self.backend.name))
+
+        if self.deprecated:
+            message.warning('{} is deprecated - {}'.format(self.name,
+                                                           self.deprecated))
 
     def _check_func(self):
         """Make sure the function parameters don't violate any rules."""
@@ -131,78 +159,47 @@ class Command:
         elif 'self' not in signature.parameters and self._instance is not None:
             raise TypeError("{} is not a class method, but instance was "
                             "given!".format(self.name[0]))
-        elif inspect.getfullargspec(self.handler).varkw is not None:
+        elif any(param.kind == inspect.Parameter.VAR_KEYWORD
+                 for param in signature.parameters.values()):
             raise TypeError("{}: functions with varkw arguments are not "
                             "supported!".format(self.name[0]))
 
-    def _get_typeconv(self, param, typ):
-        """Get a dict with a type conversion for the parameter.
+    def get_arg_info(self, param):
+        """Get an ArgInfo tuple for the given inspect.Parameter."""
+        return self._qute_args.get(param.name, ArgInfo())
 
-        Args:
-            param: The inspect.Parameter to handle.
-            typ: The type of the parameter.
-        """
-        type_conv = {}
-        if utils.is_enum(typ):
-            type_conv[param.name] = argparser.enum_getter(typ)
-        elif isinstance(typ, tuple):
-            if param.default is not inspect.Parameter.empty:
-                typ = typ + (type(param.default),)
-            type_conv[param.name] = argparser.multitype_conv(typ)
-        return type_conv
+    def get_pos_arg_info(self, pos):
+        """Get an ArgInfo tuple for the given positional parameter."""
+        name = self.pos_args[pos][0]
+        return self._qute_args.get(name, ArgInfo())
 
-    def _get_nameconv(self, param, annotation_info):
-        """Get a dict with a name conversion for the paraeter.
-
-        Args:
-            param: The inspect.Parameter to handle.
-            annotation_info: The AnnotationInfo tuple for the parameter.
-        """
-        d = {}
-        if annotation_info.name is not None:
-            d[param.name] = annotation_info.name
-        return d
-
-    def _inspect_special_param(self, param, annotation_info):
+    def _inspect_special_param(self, param):
         """Check if the given parameter is a special one.
 
         Args:
             param: The inspect.Parameter to handle.
-            annotation_info: The AnnotationInfo tuple for the parameter.
 
         Return:
             True if the parameter is special, False otherwise.
         """
-        special = annotation_info.special
-        if special == 'count':
-            if self.special_params['count'] is not None:
-                raise ValueError("Registered multiple parameters ({}/{}) as "
-                                 "count!".format(self.special_params['count'],
-                                                 param.name))
+        arg_info = self.get_arg_info(param)
+        if arg_info.count:
             if param.default is inspect.Parameter.empty:
                 raise TypeError("{}: handler has count parameter "
                                 "without default!".format(self.name))
-            self.special_params['count'] = param.name
             return True
-        elif special == 'win_id':
-            if self.special_params['win_id'] is not None:
-                raise ValueError("Registered multiple parameters ({}/{}) as "
-                                 "win_id!".format(
-                                     self.special_params['win_id'],
-                                     param.name))
-            self.special_params['win_id'] = param.name
+        elif arg_info.win_id:
             return True
-        elif special is None:
-            return False
-        else:
-            raise ValueError("{}: Invalid value '{}' for 'special' "
-                             "annotation!".format(self.name, special))
+        return False
 
     def _inspect_func(self):
         """Inspect the function to get useful informations from it.
 
         Sets instance attributes (desc, type_conv, name_conv) based on the
         informations.
+
+        Return:
+            How many user-visible arguments the command has.
         """
         signature = inspect.signature(self.handler)
         doc = inspect.getdoc(self.handler)
@@ -210,134 +207,125 @@ class Command:
             self.desc = doc.splitlines()[0].strip()
         else:
             self.desc = ""
-        if not self.ignore_args:
-            for param in signature.parameters.values():
-                annotation_info = self._parse_annotation(param)
-                if param.name == 'self':
-                    continue
-                if self._inspect_special_param(param, annotation_info):
-                    continue
-                typ = self._get_type(param, annotation_info)
-                kwargs = self._param_to_argparse_kwargs(param, annotation_info)
-                args = self._param_to_argparse_args(param, annotation_info)
-                self._type_conv.update(self._get_typeconv(param, typ))
-                self._name_conv.update(
-                    self._get_nameconv(param, annotation_info))
-                callsig = debug.format_call(
-                    self.parser.add_argument, args, kwargs,
-                    full=False)
-                log.commands.vdebug('Adding arg {} of type {} -> {}'.format(
-                    param.name, typ, callsig))
-                self.parser.add_argument(*args, **kwargs)
 
-    def _param_to_argparse_kwargs(self, param, annotation_info):
+        for param in signature.parameters.values():
+            # https://docs.python.org/3/library/inspect.html#inspect.Parameter.kind
+            # "Python has no explicit syntax for defining positional-only
+            # parameters, but many built-in and extension module functions
+            # (especially those that accept only one or two parameters) accept
+            # them."
+            assert param.kind != inspect.Parameter.POSITIONAL_ONLY
+            if param.name == 'self':
+                continue
+            if self._inspect_special_param(param):
+                continue
+            if (param.kind == inspect.Parameter.KEYWORD_ONLY and
+                    param.default is inspect.Parameter.empty):
+                raise TypeError("{}: handler has keyword only argument {!r} "
+                                "without default!".format(
+                                    self.name, param.name))
+            typ = self._get_type(param)
+            is_bool = typ is bool
+            kwargs = self._param_to_argparse_kwargs(param, is_bool)
+            args = self._param_to_argparse_args(param, is_bool)
+            callsig = debug_utils.format_call(self.parser.add_argument, args,
+                                              kwargs, full=False)
+            log.commands.vdebug('Adding arg {} of type {} -> {}'.format(
+                param.name, typ, callsig))
+            self.parser.add_argument(*args, **kwargs)
+        return signature.parameters.values()
+
+    def _param_to_argparse_kwargs(self, param, is_bool):
         """Get argparse keyword arguments for a parameter.
 
         Args:
             param: The inspect.Parameter object to get the args for.
-            annotation_info: An AnnotationInfo tuple for the parameter.
+            is_bool: Whether the parameter is a boolean.
 
         Return:
             A kwargs dict.
         """
         kwargs = {}
-        typ = self._get_type(param, annotation_info)
 
         try:
             kwargs['help'] = self.docparser.arg_descs[param.name]
         except KeyError:
             pass
 
-        if isinstance(typ, tuple):
-            pass
-        elif utils.is_enum(typ):
-            kwargs['choices'] = [e.name.replace('_', '-') for e in typ]
-            kwargs['metavar'] = param.name
-        elif typ is bool:
+        kwargs['dest'] = param.name
+
+        arg_info = self.get_arg_info(param)
+
+        if is_bool:
             kwargs['action'] = 'store_true'
-        elif typ is not None:
-            kwargs['type'] = typ
+        else:
+            if arg_info.metavar is not None:
+                kwargs['metavar'] = arg_info.metavar
+            else:
+                kwargs['metavar'] = argparser.arg_name(param.name)
 
         if param.kind == inspect.Parameter.VAR_POSITIONAL:
-            kwargs['nargs'] = '+'
+            kwargs['nargs'] = '*' if self._star_args_optional else '+'
         elif param.kind == inspect.Parameter.KEYWORD_ONLY:
             kwargs['default'] = param.default
-        elif typ is not bool and param.default is not inspect.Parameter.empty:
+        elif not is_bool and param.default is not inspect.Parameter.empty:
             kwargs['default'] = param.default
             kwargs['nargs'] = '?'
-        kwargs.update(annotation_info.kwargs)
         return kwargs
 
-    def _param_to_argparse_args(self, param, annotation_info):
+    def _param_to_argparse_args(self, param, is_bool):
         """Get argparse positional arguments for a parameter.
 
         Args:
             param: The inspect.Parameter object to get the args for.
-            annotation_info: An AnnotationInfo tuple for the parameter.
+            is_bool: Whether the parameter is a boolean.
 
         Return:
             A list of args.
         """
         args = []
-        name = annotation_info.name or param.name
-        shortname = annotation_info.flag or param.name[0]
+        name = argparser.arg_name(param.name)
+        arg_info = self.get_arg_info(param)
+
+        if arg_info.flag is not None:
+            shortname = arg_info.flag
+        else:
+            shortname = name[0]
+
         if len(shortname) != 1:
             raise ValueError("Flag '{}' of parameter {} (command {}) must be "
                              "exactly 1 char!".format(shortname, name,
                                                       self.name))
-        typ = self._get_type(param, annotation_info)
-        if typ is bool or param.kind == inspect.Parameter.KEYWORD_ONLY:
+        if is_bool or param.kind == inspect.Parameter.KEYWORD_ONLY:
             long_flag = '--{}'.format(name)
             short_flag = '-{}'.format(shortname)
             args.append(long_flag)
             args.append(short_flag)
             self.opt_args[param.name] = long_flag, short_flag
-            if param.kind == inspect.Parameter.KEYWORD_ONLY:
-                self.flags_with_args.append(param.name)
+            if not is_bool:
+                self.flags_with_args += [short_flag, long_flag]
         else:
-            args.append(name)
-            self.pos_args.append((param.name, name))
+            if not arg_info.hide:
+                self.pos_args.append((param.name, name))
         return args
 
-    def _parse_annotation(self, param):
-        """Get argparse arguments and type from a parameter annotation.
-
-        Args:
-            param: A inspect.Parameter instance.
-
-        Return:
-            An AnnotationInfo namedtuple.
-                kwargs: A dict of keyword args to add to the
-                        argparse.ArgumentParser.add_argument call.
-                typ: The type to use for this argument.
-                flag: The short name/flag if overridden.
-                name: The long name if overridden.
-        """
-        info = {'kwargs': {}, 'type': None, 'flag': None, 'name': None,
-                'special': None}
-        if param.annotation is not inspect.Parameter.empty:
-            log.commands.vdebug("Parsing annotation {}".format(
-                param.annotation))
-            for field in ('type', 'flag', 'name', 'special'):
-                if field in param.annotation:
-                    info[field] = param.annotation[field]
-            if 'nargs' in param.annotation:
-                info['kwargs'] = {'nargs': param.annotation['nargs']}
-        return self.AnnotationInfo(**info)
-
-    def _get_type(self, param, annotation_info):
+    def _get_type(self, param):
         """Get the type of an argument from its default value or annotation.
 
         Args:
             param: The inspect.Parameter to look at.
-            annotation_info: An AnnotationInfo tuple which overrides the type.
         """
-        if annotation_info.type is not None:
-            return annotation_info.type
-        elif param.default is None or param.default is inspect.Parameter.empty:
+        arginfo = self.get_arg_info(param)
+        if param.annotation is not inspect.Parameter.empty:
+            return param.annotation
+        elif param.default not in [None, inspect.Parameter.empty]:
+            return type(param.default)
+        elif arginfo.count or arginfo.win_id or param.kind in [
+                inspect.Parameter.VAR_POSITIONAL,
+                inspect.Parameter.VAR_KEYWORD]:
             return None
         else:
-            return type(param.default)
+            return str
 
     def _get_self_arg(self, win_id, param, args):
         """Get the self argument for a function call.
@@ -348,10 +336,17 @@ class Command:
             args: The positional argument list. Gets modified directly.
         """
         assert param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD
-        if self._scope is not 'window':
+        if self._scope == 'global':
+            tab_id = None
             win_id = None
-        obj = objreg.get(self._instance, scope=self._scope,
-                         window=win_id)
+        elif self._scope == 'tab':
+            tab_id = 'current'
+        elif self._scope == 'window':
+            tab_id = None
+        else:
+            raise ValueError("Invalid scope {}!".format(self._scope))
+        obj = objreg.get(self._instance, scope=self._scope, window=win_id,
+                         tab=tab_id)
         args.append(obj)
 
     def _get_count_arg(self, param, args, kwargs):
@@ -391,19 +386,43 @@ class Command:
             raise TypeError("{}: invalid parameter type {} for argument "
                             "{!r}!".format(self.name, param.kind, param.name))
 
-    def _get_param_name_and_value(self, param):
-        """Get the converted name and value for an inspect.Parameter."""
-        name = self._name_conv.get(param.name, param.name)
-        value = getattr(self.namespace, name)
-        if param.name in self._type_conv:
-            # We convert enum types after getting the values from
-            # argparse, because argparse's choices argument is
-            # processed after type conversation, which is not what we
-            # want.
-            value = self._type_conv[param.name](value)
-        return name, value
+    def _get_param_value(self, param):
+        """Get the converted value for an inspect.Parameter."""
+        value = getattr(self.namespace, param.name)
+        typ = self._get_type(param)
 
-    def _get_call_args(self, win_id):  # noqa
+        if isinstance(typ, tuple):
+            raise TypeError("{}: Legacy tuple type annotation!".format(
+                self.name))
+        elif type(typ) is type(typing.Union):  # noqa: E721
+            # this is... slightly evil, I know
+            # We also can't use isinstance here because typing.Union doesn't
+            # support that.
+            # pylint: disable=no-member,useless-suppression
+            try:
+                types = list(typ.__args__)
+            except AttributeError:
+                # Older Python 3.5 patch versions
+                types = list(typ.__union_params__)
+            # pylint: enable=no-member,useless-suppression
+            if param.default is not inspect.Parameter.empty:
+                types.append(type(param.default))
+            choices = self.get_arg_info(param).choices
+            value = argparser.multitype_conv(param, types, value,
+                                             str_choices=choices)
+        elif typ is str:
+            choices = self.get_arg_info(param).choices
+            value = argparser.type_conv(param, typ, value, str_choices=choices)
+        elif typ is bool:  # no type conversion for flags
+            assert isinstance(value, bool)
+        elif typ is None:
+            pass
+        else:
+            value = argparser.type_conv(param, typ, value)
+
+        return value
+
+    def _get_call_args(self, win_id):
         """Get arguments for a function call.
 
         Args:
@@ -412,38 +431,33 @@ class Command:
         Return:
             An (args, kwargs) tuple.
         """
-
         args = []
         kwargs = {}
         signature = inspect.signature(self.handler)
 
-        if self.ignore_args:
-            if self._instance is not None:
-                param = list(signature.parameters.values())[0]
-                self._get_self_arg(win_id, param, args)
-            return args, kwargs
-
         for i, param in enumerate(signature.parameters.values()):
+            arg_info = self.get_arg_info(param)
             if i == 0 and self._instance is not None:
                 # Special case for 'self'.
                 self._get_self_arg(win_id, param, args)
                 continue
-            elif param.name == self.special_params['count']:
+            elif arg_info.count:
                 # Special case for count parameter.
                 self._get_count_arg(param, args, kwargs)
                 continue
-            elif param.name == self.special_params['win_id']:
+            # elif arg_info.win_id:
+            elif arg_info.win_id:
                 # Special case for win_id parameter.
                 self._get_win_id_arg(win_id, param, args, kwargs)
                 continue
-            name, value = self._get_param_name_and_value(param)
+            value = self._get_param_value(param)
             if param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
                 args.append(value)
             elif param.kind == inspect.Parameter.VAR_POSITIONAL:
                 if value is not None:
                     args += value
             elif param.kind == inspect.Parameter.KEYWORD_ONLY:
-                kwargs[name] = value
+                kwargs[param.name] = value
             else:
                 raise TypeError("{}: Invalid parameter type {} for argument "
                                 "'{}'!".format(
@@ -463,21 +477,41 @@ class Command:
         dbgout = ["command called:", self.name]
         if args:
             dbgout.append(str(args))
+        elif args is None:
+            args = []
+
         if count is not None:
             dbgout.append("(count={})".format(count))
         log.commands.debug(' '.join(dbgout))
         try:
             self.namespace = self.parser.parse_args(args)
         except argparser.ArgumentParserError as e:
-            message.error(win_id, '{}: {}'.format(self.name, e))
+            message.error('{}: {}'.format(self.name, e),
+                          stack=traceback.format_exc())
             return
         except argparser.ArgumentParserExit as e:
             log.commands.debug("argparser exited with status {}: {}".format(
                 e.status, e))
             return
         self._count = count
-        posargs, kwargs = self._get_call_args(win_id)
         self._check_prerequisites(win_id)
+        posargs, kwargs = self._get_call_args(win_id)
         log.commands.debug('Calling {}'.format(
-            debug.format_call(self.handler, posargs, kwargs)))
+            debug_utils.format_call(self.handler, posargs, kwargs)))
         self.handler(*posargs, **kwargs)
+
+    def validate_mode(self, mode):
+        """Raise cmdexc.PrerequisitesError unless allowed in the given mode.
+
+        Args:
+            mode: The usertypes.KeyMode to check.
+        """
+        if mode not in self.modes:
+            mode_names = '/'.join(sorted(m.name for m in self.modes))
+            raise cmdexc.PrerequisitesError(
+                "{}: This command is only allowed in {} mode, not {}.".format(
+                    self.name, mode_names, mode.name))
+
+    def takes_count(self):
+        """Return true iff this command can take a count argument."""
+        return any(arg.count for arg in self._qute_args)
