@@ -1,6 +1,6 @@
 # vim: ft=python fileencoding=utf-8 sts=4 sw=4 et:
 
-# Copyright 2014-2015 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
+# Copyright 2014-2017 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
 #
 # This file is part of qutebrowser.
 #
@@ -17,23 +17,42 @@
 # You should have received a copy of the GNU General Public License
 # along with qutebrowser.  If not, see <http://www.gnu.org/licenses/>.
 
-"""Mode manager singleton which handles the current keyboard mode.
-
-Module attributes:
-    manager: The ModeManager instance.
-"""
+"""Mode manager singleton which handles the current keyboard mode."""
 
 import functools
 
-from PyQt5.QtGui import QWindow
-from PyQt5.QtCore import pyqtSignal, Qt, QObject, QEvent
+import attr
+from PyQt5.QtCore import pyqtSlot, pyqtSignal, Qt, QObject, QEvent
 from PyQt5.QtWidgets import QApplication
-from PyQt5.QtWebKitWidgets import QWebView
 
 from qutebrowser.keyinput import modeparsers, keyparser
 from qutebrowser.config import config
 from qutebrowser.commands import cmdexc, cmdutils
 from qutebrowser.utils import usertypes, log, objreg, utils
+from qutebrowser.misc import objects
+
+
+@attr.s(frozen=True)
+class KeyEvent:
+
+    """A small wrapper over a QKeyEvent storing its data.
+
+    This is needed because Qt apparently mutates existing events with new data.
+    It doesn't store the modifiers because they can be different for a key
+    press/release.
+
+    Attributes:
+        key: A Qt.Key member (QKeyEvent::key).
+        text: A string (QKeyEvent::text).
+    """
+
+    key = attr.ib()
+    text = attr.ib()
+
+    @classmethod
+    def from_event(cls, event):
+        """Initialize a KeyEvent from a QKeyEvent."""
+        return cls(event.key(), event.text())
 
 
 class NotInModeError(Exception):
@@ -43,7 +62,7 @@ class NotInModeError(Exception):
 
 def init(win_id, parent):
     """Initialize the mode manager and the keyparsers for the given win_id."""
-    KM = usertypes.KeyMode  # pylint: disable=invalid-name
+    KM = usertypes.KeyMode  # noqa: N801,N806 pylint: disable=invalid-name
     modeman = ModeManager(win_id, parent)
     objreg.register('mode-manager', modeman, scope='window', window=win_id)
     keyparsers = {
@@ -56,85 +75,38 @@ def init(win_id, parent):
         KM.prompt: keyparser.PassthroughKeyParser(win_id, 'prompt', modeman,
                                                   warn=False),
         KM.yesno: modeparsers.PromptKeyParser(win_id, modeman),
+        KM.caret: modeparsers.CaretKeyParser(win_id, modeman),
+        KM.set_mark: modeparsers.RegisterKeyParser(win_id, KM.set_mark,
+                                                   modeman),
+        KM.jump_mark: modeparsers.RegisterKeyParser(win_id, KM.jump_mark,
+                                                    modeman),
+        KM.record_macro: modeparsers.RegisterKeyParser(win_id, KM.record_macro,
+                                                       modeman),
+        KM.run_macro: modeparsers.RegisterKeyParser(win_id, KM.run_macro,
+                                                    modeman),
     }
     objreg.register('keyparsers', keyparsers, scope='window', window=win_id)
     modeman.destroyed.connect(
         functools.partial(objreg.delete, 'keyparsers', scope='window',
                           window=win_id))
-    modeman.register(KM.normal, keyparsers[KM.normal].handle)
-    modeman.register(KM.hint, keyparsers[KM.hint].handle)
-    modeman.register(KM.insert, keyparsers[KM.insert].handle, passthrough=True)
-    modeman.register(KM.passthrough, keyparsers[KM.passthrough].handle,
-                     passthrough=True)
-    modeman.register(KM.command, keyparsers[KM.command].handle,
-                     passthrough=True)
-    modeman.register(KM.prompt, keyparsers[KM.prompt].handle, passthrough=True)
-    modeman.register(KM.yesno, keyparsers[KM.yesno].handle)
+    for mode, parser in keyparsers.items():
+        modeman.register(mode, parser)
     return modeman
 
 
-def _get_modeman(win_id):
+def instance(win_id):
     """Get a modemanager object."""
     return objreg.get('mode-manager', scope='window', window=win_id)
 
 
 def enter(win_id, mode, reason=None, only_if_normal=False):
     """Enter the mode 'mode'."""
-    _get_modeman(win_id).enter(mode, reason, only_if_normal)
+    instance(win_id).enter(mode, reason, only_if_normal)
 
 
-def leave(win_id, mode, reason=None):
+def leave(win_id, mode, reason=None, *, maybe=False):
     """Leave the mode 'mode'."""
-    _get_modeman(win_id).leave(mode, reason)
-
-
-def maybe_leave(win_id, mode, reason=None):
-    """Convenience method to leave 'mode' without exceptions."""
-    try:
-        _get_modeman(win_id).leave(mode, reason)
-    except NotInModeError as e:
-        # This is rather likely to happen, so we only log to debug log.
-        log.modes.debug("{} (leave reason: {})".format(e, reason))
-
-
-class EventFilter(QObject):
-
-    """Event filter which passes the event to the corrent ModeManager."""
-
-    def __init__(self, parent=None):
-        super().__init__(parent)
-        self._activated = True
-
-    def eventFilter(self, obj, event):
-        """Forward events to the correct modeman."""
-        try:
-            if not self._activated:
-                return False
-            if event.type() not in [QEvent.KeyPress, QEvent.KeyRelease]:
-                # We're not interested in non-key-events so we pass them
-                # through.
-                return False
-            if not isinstance(obj, QWindow):
-                # We already handled this same event at some point earlier, so
-                # we're not interested in it anymore.
-                return False
-            if (QApplication.instance().activeWindow() not in
-                    objreg.window_registry.values()):
-                # Some other window (print dialog, etc.) is focused so we pass
-                # the event through.
-                return False
-            try:
-                modeman = objreg.get('mode-manager', scope='window',
-                                     window='current')
-                return modeman.eventFilter(event)
-            except objreg.RegistryUnavailableError:
-                # No window available yet, or not a MainWindow
-                return False
-        except:
-            # If there is an exception in here and we leave the eventfilter
-            # activated, we'll get an infinite loop and a stack overflow.
-            self._activated = False
-            raise
+    instance(win_id).leave(mode, reason, maybe=maybe)
 
 
 class ModeManager(QObject):
@@ -142,12 +114,11 @@ class ModeManager(QObject):
     """Manager for keyboard modes.
 
     Attributes:
-        passthrough: A list of modes in which to pass through events.
         mode: The mode we're currently in.
         _win_id: The window ID of this ModeManager
-        _handlers: A dictionary of modes and their handlers.
+        _parsers: A dictionary of modes and their keyparsers.
         _forward_unbound_keys: If we should forward unbound keys.
-        _releaseevents_to_pass: A list of keys where the keyPressEvent was
+        _releaseevents_to_pass: A set of KeyEvents where the keyPressEvent was
                                 passed through, so the release event should as
                                 well.
 
@@ -167,17 +138,12 @@ class ModeManager(QObject):
     def __init__(self, win_id, parent=None):
         super().__init__(parent)
         self._win_id = win_id
-        self._handlers = {}
-        self.passthrough = []
+        self._parsers = {}
         self.mode = usertypes.KeyMode.normal
-        self._releaseevents_to_pass = []
-        self._forward_unbound_keys = config.get(
-            'input', 'forward-unbound-keys')
-        objreg.get('config').changed.connect(self.set_forward_unbound_keys)
+        self._releaseevents_to_pass = set()
 
     def __repr__(self):
-        return utils.get_repr(self, mode=self.mode,
-                              passthrough=self.passthrough)
+        return utils.get_repr(self, mode=self.mode)
 
     def _eventFilter_keypress(self, event):
         """Handle filtering of KeyPress events.
@@ -189,37 +155,37 @@ class ModeManager(QObject):
             True if event should be filtered, False otherwise.
         """
         curmode = self.mode
-        handler = self._handlers[curmode]
+        parser = self._parsers[curmode]
         if curmode != usertypes.KeyMode.insert:
-            log.modes.debug("got keypress in mode {} - calling handler "
-                            "{}".format(curmode, utils.qualname(handler)))
-        handled = handler(event) if handler is not None else False
+            log.modes.debug("got keypress in mode {} - delegating to "
+                            "{}".format(curmode, utils.qualname(parser)))
+        handled = parser.handle(event)
 
-        is_non_alnum = bool(event.modifiers()) or not event.text().strip()
-        focus_widget = QApplication.instance().focusWidget()
-        is_tab = event.key() in (Qt.Key_Tab, Qt.Key_Backtab)
+        is_non_alnum = (
+            event.modifiers() not in [Qt.NoModifier, Qt.ShiftModifier] or
+            not event.text().strip())
+
+        forward_unbound_keys = config.val.input.forward_unbound_keys
 
         if handled:
             filter_this = True
-        elif is_tab and not isinstance(focus_widget, QWebView):
-            filter_this = True
-        elif (curmode in self.passthrough or
-                self._forward_unbound_keys == 'all' or
-                (self._forward_unbound_keys == 'auto' and is_non_alnum)):
+        elif (parser.passthrough or forward_unbound_keys == 'all' or
+              (forward_unbound_keys == 'auto' and is_non_alnum)):
             filter_this = False
         else:
             filter_this = True
 
         if not filter_this:
-            self._releaseevents_to_pass.append(event)
+            self._releaseevents_to_pass.add(KeyEvent.from_event(event))
 
         if curmode != usertypes.KeyMode.insert:
-            log.modes.debug("handled: {}, forward-unbound-keys: {}, "
-                            "passthrough: {}, is_non_alnum: {}, is_tab {} --> "
+            focus_widget = QApplication.instance().focusWidget()
+            log.modes.debug("handled: {}, forward_unbound_keys: {}, "
+                            "passthrough: {}, is_non_alnum: {} --> "
                             "filter: {} (focused: {!r})".format(
-                                handled, self._forward_unbound_keys,
-                                curmode in self.passthrough, is_non_alnum,
-                                is_tab, filter_this, focus_widget))
+                                handled, forward_unbound_keys,
+                                parser.passthrough, is_non_alnum, filter_this,
+                                focus_widget))
         return filter_this
 
     def _eventFilter_keyrelease(self, event):
@@ -232,10 +198,9 @@ class ModeManager(QObject):
             True if event should be filtered, False otherwise.
         """
         # handle like matching KeyPress
-        if event in self._releaseevents_to_pass:
-            # remove all occurences
-            self._releaseevents_to_pass = [
-                e for e in self._releaseevents_to_pass if e != event]
+        keyevent = KeyEvent.from_event(event)
+        if keyevent in self._releaseevents_to_pass:
+            self._releaseevents_to_pass.remove(keyevent)
             filter_this = False
         else:
             filter_this = True
@@ -243,20 +208,17 @@ class ModeManager(QObject):
             log.modes.debug("filter: {}".format(filter_this))
         return filter_this
 
-    def register(self, mode, handler, passthrough=False):
+    def register(self, mode, parser):
         """Register a new mode.
 
         Args:
             mode: The name of the mode.
-            handler: Handler for keyPressEvents.
-            passthrough: Whether to pass keybindings in this mode through to
-                         the widgets.
+            parser: The KeyParser which should be used.
         """
-        if not isinstance(mode, usertypes.KeyMode):
-            raise TypeError("Mode {} is no KeyMode member!".format(mode))
-        self._handlers[mode] = handler
-        if passthrough:
-            self.passthrough.append(mode)
+        assert isinstance(mode, usertypes.KeyMode)
+        assert parser is not None
+        self._parsers[mode] = parser
+        parser.request_leave.connect(self.leave)
 
     def enter(self, mode, reason=None, only_if_normal=False):
         """Enter a new mode.
@@ -270,8 +232,8 @@ class ModeManager(QObject):
             raise TypeError("Mode {} is no KeyMode member!".format(mode))
         log.modes.debug("Entering mode {}{}".format(
             mode, '' if reason is None else ' (reason: {})'.format(reason)))
-        if mode not in self._handlers:
-            raise ValueError("No handler for mode {}".format(mode))
+        if mode not in self._parsers:
+            raise ValueError("No keyparser for mode {}".format(mode))
         prompt_modes = (usertypes.KeyMode.prompt, usertypes.KeyMode.yesno)
         if self.mode == mode or (self.mode in prompt_modes and
                                  mode in prompt_modes):
@@ -289,7 +251,7 @@ class ModeManager(QObject):
         self.mode = mode
         self.entered.emit(mode, self._win_id)
 
-    @cmdutils.register(instance='mode-manager', hide=True, scope='window')
+    @cmdutils.register(instance='mode-manager', scope='window')
     def enter_mode(self, mode):
         """Enter a key mode.
 
@@ -300,37 +262,52 @@ class ModeManager(QObject):
             m = usertypes.KeyMode[mode]
         except KeyError:
             raise cmdexc.CommandError("Mode {} does not exist!".format(mode))
+
+        if m in [usertypes.KeyMode.hint, usertypes.KeyMode.command,
+                 usertypes.KeyMode.yesno, usertypes.KeyMode.prompt]:
+            raise cmdexc.CommandError(
+                "Mode {} can't be entered manually!".format(mode))
+        elif (m == usertypes.KeyMode.caret and
+              objects.backend == usertypes.Backend.QtWebEngine):
+            raise cmdexc.CommandError("Caret mode is not supported with "
+                                      "QtWebEngine yet.")
+
         self.enter(m, 'command')
 
-    def leave(self, mode, reason=None):
+    @pyqtSlot(usertypes.KeyMode, str, bool)
+    def leave(self, mode, reason=None, maybe=False):
         """Leave a key mode.
 
         Args:
-            mode: The name of the mode to leave.
+            mode: The mode to leave as a usertypes.KeyMode member.
             reason: Why the mode was left.
+            maybe: If set, ignore the request if we're not in that mode.
         """
         if self.mode != mode:
-            raise NotInModeError("Not in mode {}!".format(mode))
+            if maybe:
+                log.modes.debug("Ignoring leave request for {} (reason {}) as "
+                                "we're in mode {}".format(
+                                    mode, reason, self.mode))
+                return
+            else:
+                raise NotInModeError("Not in mode {}!".format(mode))
+
         log.modes.debug("Leaving mode {}{}".format(
             mode, '' if reason is None else ' (reason: {})'.format(reason)))
+        # leaving a mode implies clearing keychain, see
+        # https://github.com/qutebrowser/qutebrowser/issues/1805
+        self.clear_keychain()
         self.mode = usertypes.KeyMode.normal
         self.left.emit(mode, self.mode, self._win_id)
 
     @cmdutils.register(instance='mode-manager', name='leave-mode',
-                       not_modes=[usertypes.KeyMode.normal], hide=True,
-                       scope='window')
+                       not_modes=[usertypes.KeyMode.normal], scope='window')
     def leave_current_mode(self):
         """Leave the mode we're currently in."""
         if self.mode == usertypes.KeyMode.normal:
             raise ValueError("Can't leave normal mode!")
         self.leave(self.mode, 'leave current')
 
-    @config.change_filter('input', 'forward-unbound-keys')
-    def set_forward_unbound_keys(self):
-        """Update local setting when config changed."""
-        self._forward_unbound_keys = config.get(
-            'input', 'forward-unbound-keys')
-
     def eventFilter(self, event):
         """Filter all events based on the currently set mode.
 
@@ -345,7 +322,15 @@ class ModeManager(QObject):
         if self.mode is None:
             # We got events before mode is set, so just pass them through.
             return False
-        if event.type() == QEvent.KeyPress:
-            return self._eventFilter_keypress(event)
-        else:
-            return self._eventFilter_keyrelease(event)
+
+        handlers = {
+            QEvent.KeyPress: self._eventFilter_keypress,
+            QEvent.KeyRelease: self._eventFilter_keyrelease,
+        }
+        handler = handlers[event.type()]
+        return handler(event)
+
+    @cmdutils.register(instance='mode-manager', scope='window')
+    def clear_keychain(self):
+        """Clear the currently entered key chain."""
+        self._parsers[self.mode].clear_keystring()
