@@ -1,6 +1,6 @@
 # vim: ft=python fileencoding=utf-8 sts=4 sw=4 et:
 
-# Copyright 2014-2015 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
+# Copyright 2014-2017 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
 #
 # This file is part of qutebrowser.
 #
@@ -17,26 +17,33 @@
 # You should have received a copy of the GNU General Public License
 # along with qutebrowser.  If not, see <http://www.gnu.org/licenses/>.
 
-# pylint: disable=broad-except
-
 """The dialog which gets shown when qutebrowser crashes."""
 
 import re
+import os
 import sys
 import html
 import getpass
+import fnmatch
 import traceback
+import datetime
+import enum
 
-from PyQt5.QtCore import pyqtSlot, Qt, QSize, qVersion
+import pkg_resources
+from PyQt5.QtCore import pyqtSlot, Qt, QSize
 from PyQt5.QtWidgets import (QDialog, QLabel, QTextEdit, QPushButton,
                              QVBoxLayout, QHBoxLayout, QCheckBox,
-                             QDialogButtonBox, QMessageBox)
+                             QDialogButtonBox, QApplication, QMessageBox)
 
 import qutebrowser
-from qutebrowser.utils import version, log, utils, objreg, qtutils
-from qutebrowser.misc import miscwidgets
-from qutebrowser.browser.network import pastebin
-from qutebrowser.config import config
+from qutebrowser.utils import version, log, utils, objreg
+from qutebrowser.misc import (miscwidgets, autoupdate, msgbox, httpclient,
+                              pastebin)
+from qutebrowser.config import config, configfiles
+
+
+Result = enum.IntEnum('Result', ['restore', 'no_restore'],
+                      start=QDialog.Accepted + 1)
 
 
 def parse_fatal_stacktrace(text):
@@ -50,46 +57,34 @@ def parse_fatal_stacktrace(text):
         element being the first stacktrace frame.
     """
     lines = [
-        r'Fatal Python error: (.*)',
+        r'(?P<type>Fatal Python error|Windows fatal exception): (?P<msg>.*)',
         r' *',
         r'(Current )?[Tt]hread [^ ]* \(most recent call first\): *',
-        r'  File ".*", line \d+ in (.*)',
+        r'  File ".*", line \d+ in (?P<func>.*)',
     ]
-    m = re.match('\n'.join(lines), text)
+    m = re.search('\n'.join(lines), text)
     if m is None:
         # We got some invalid text.
         return ('', '')
     else:
-        return (m.group(1), m.group(3))
+        msg = m.group('msg')
+        typ = m.group('type')
+        func = m.group('func')
+        if typ == 'Windows fatal exception':
+            msg = 'Windows ' + msg
+        return msg, func
 
 
-def get_fatal_crash_dialog(debug, data):
-    """Get a fatal crash dialog based on a crash log.
-
-    If the crash is a segfault in qt_mainloop and we're on an old Qt version
-    this is a simple error dialog which lets the user know they should upgrade
-    if possible.
-
-    If it's anything else, it's a normal FatalCrashDialog with the possibility
-    to report the crash.
-
-    Args:
-        debug: Whether the debug flag (--debug) was given.
-        data: The crash log data.
-    """
-    errtype, frame = parse_fatal_stacktrace(data)
-    if (qtutils.version_check('5.4') or errtype != 'Segmentation fault' or
-            frame != 'qt_mainloop'):
-        return FatalCrashDialog(debug, data)
-    else:
-        title = "qutebrowser was restarted after a fatal crash!"
-        text = ("<b>qutebrowser was restarted after a fatal crash!</b><br/>"
-                "Unfortunately, this crash occured in Qt (the library "
-                "qutebrowser uses), and your version ({}) is outdated - "
-                "Qt 5.4 or later is recommended. Unfortuntately Debian and "
-                "Ubuntu don't ship a newer version (yet?)...".format(
-                    qVersion()))
-        return QMessageBox(QMessageBox.Critical, title, text, QMessageBox.Ok)
+def _get_environment_vars():
+    """Gather environment variables for the crash info."""
+    masks = ('DESKTOP_SESSION', 'DE', 'QT_*', 'PYTHON*', 'LC_*', 'LANG',
+             'XDG_*', 'QUTE_*', 'PATH')
+    info = []
+    for key, value in os.environ.items():
+        for m in masks:
+            if fnmatch.fnmatch(key, m):
+                info.append('{} = {}'.format(key, value))
+    return '\n'.join(sorted(info))
 
 
 class _CrashDialog(QDialog):
@@ -105,6 +100,7 @@ class _CrashDialog(QDialog):
         _url: Pastebin URL QLabel.
         _crash_info: A list of tuples with title and crash information.
         _paste_client: A PastebinClient instance to use.
+        _pypi_client: A PyPIVersionClient instance to use.
         _paste_text: The text to pastebin.
     """
 
@@ -126,25 +122,12 @@ class _CrashDialog(QDialog):
         self.setWindowTitle("Whoops!")
         self.resize(QSize(640, 600))
         self._vbox = QVBoxLayout(self)
-        self._paste_client = pastebin.PastebinClient(self)
+        http_client = httpclient.HTTPClient()
+        self._paste_client = pastebin.PastebinClient(http_client, self)
+        self._pypi_client = autoupdate.PyPIVersionClient(self)
         self._init_text()
 
-        contact = QLabel("I'd like to be able to follow up with you, to keep "
-                         "you posted on the status of this crash and get more "
-                         "information if I need it - how can I contact you?",
-                         wordWrap=True)
-        self._vbox.addWidget(contact)
-        self._contact = QTextEdit(tabChangesFocus=True, acceptRichText=False)
-        try:
-            state = objreg.get('state-config')
-            try:
-                self._contact.setPlainText(state['general']['contact-info'])
-            except KeyError:
-                self._contact.setPlaceholderText("Mail or IRC nickname")
-        except Exception:
-            log.misc.exception("Failed to get contact information!")
-            self._contact.setPlaceholderText("Mail or IRC nickname")
-        self._vbox.addWidget(self._contact, 2)
+        self._init_contact_input()
 
         info = QLabel("What were you doing when this crash/bug happened?")
         self._vbox.addWidget(info)
@@ -178,10 +161,31 @@ class _CrashDialog(QDialog):
     def __repr__(self):
         return utils.get_repr(self)
 
+    def _init_contact_input(self):
+        """Initialize the widget asking for contact info."""
+        contact = QLabel("I'd like to be able to follow up with you, to keep "
+                         "you posted on the status of this crash and get more "
+                         "information if I need it - how can I contact you?",
+                         wordWrap=True)
+        self._vbox.addWidget(contact)
+        self._contact = QTextEdit(tabChangesFocus=True, acceptRichText=False)
+        try:
+            try:
+                info = configfiles.state['general']['contact-info']
+            except KeyError:
+                self._contact.setPlaceholderText("Mail or IRC nickname")
+            else:
+                self._contact.setPlainText(info)
+        except Exception:
+            log.misc.exception("Failed to get contact information!")
+            self._contact.setPlaceholderText("Mail or IRC nickname")
+        self._vbox.addWidget(self._contact, 2)
+
     def _init_text(self):
         """Initialize the main text to be displayed on an exception.
 
-        Should be extended by superclass to set the actual text."""
+        Should be extended by subclasses to set the actual text.
+        """
         self._lbl = QLabel(wordWrap=True, openExternalLinks=True,
                            textInteractionFlags=Qt.LinksAccessibleByMouse)
         self._vbox.addWidget(self._lbl)
@@ -206,8 +210,8 @@ class _CrashDialog(QDialog):
     def _init_info_text(self):
         """Add an info text encouraging the user to report crashes."""
         info_label = QLabel("<br/><b>Note that without your help, I can't fix "
-                            "the bug you encountered.<br/>I read and respond "
-                            "to all crash reports!</b>", wordWrap=True)
+                            "the bug you encountered.</b>",
+                            wordWrap=True)
         self._vbox.addWidget(info_label)
 
     def _gather_crash_info(self):
@@ -218,15 +222,27 @@ class _CrashDialog(QDialog):
             cmdhist: A list with the command history (as strings)
             exc: An exception tuple (type, value, traceback)
         """
+        try:
+            application = QApplication.instance()
+            launch_time = application.launch_time.ctime()
+            crash_time = datetime.datetime.now().ctime()
+            text = 'Launch: {}\nCrash: {}'.format(launch_time, crash_time)
+            self._crash_info.append(('Timestamps', text))
+        except Exception:
+            self._crash_info.append(("Launch time", traceback.format_exc()))
         try:
             self._crash_info.append(("Version info", version.version()))
         except Exception:
             self._crash_info.append(("Version info", traceback.format_exc()))
         try:
-            conf = objreg.get('config')
-            self._crash_info.append(("Config", conf.dump_userconfig()))
+            self._crash_info.append(("Config",
+                                     config.instance.dump_userconfig()))
         except Exception:
             self._crash_info.append(("Config", traceback.format_exc()))
+        try:
+            self._crash_info.append(("Environment", _get_environment_vars()))
+        except Exception:
+            self._crash_info.append(("Environment", traceback.format_exc()))
 
     def _set_crash_info(self):
         """Set/update the crash info."""
@@ -251,8 +267,8 @@ class _CrashDialog(QDialog):
     def _get_paste_title(self):
         """Get a title for the paste."""
         desc = self._get_paste_title_desc()
-        title = "qute {} {}".format(
-            qutebrowser.__version__, self._get_error_type())
+        title = "qute {} {}".format(qutebrowser.__version__,
+                                    self._get_error_type())
         if desc:
             title += ' {}'.format(desc)
         return title
@@ -260,8 +276,8 @@ class _CrashDialog(QDialog):
     def _save_contact_info(self):
         """Save the contact info to disk."""
         try:
-            state = objreg.get('state-config')
-            state['general']['contact-info'] = self._contact.toPlainText()
+            info = self._contact.toPlainText()
+            configfiles.state['general']['contact-info'] = info
         except Exception:
             log.misc.exception("Failed to save contact information!")
 
@@ -295,10 +311,17 @@ class _CrashDialog(QDialog):
         self._btn_report.setEnabled(False)
         self._btn_cancel.setEnabled(False)
         self._btn_report.setText("Reporting...")
-        self._paste_client.success.connect(self.finish)
+        self._paste_client.success.connect(self.on_paste_success)
         self._paste_client.error.connect(self.show_error)
         self.report()
 
+    @pyqtSlot()
+    def on_paste_success(self):
+        """Get the newest version from PyPI when the paste is done."""
+        self._pypi_client.success.connect(self.on_version_success)
+        self._pypi_client.error.connect(self.on_version_error)
+        self._pypi_client.get_version()
+
     @pyqtSlot(str)
     def show_error(self, text):
         """Show a paste error dialog.
@@ -310,6 +333,40 @@ class _CrashDialog(QDialog):
         error_dlg.finished.connect(self.finish)
         error_dlg.show()
 
+    @pyqtSlot(str)
+    def on_version_success(self, newest):
+        """Called when the version was obtained from self._pypi_client.
+
+        Args:
+            newest: The newest version as a string.
+        """
+        new_version = pkg_resources.parse_version(newest)
+        cur_version = pkg_resources.parse_version(qutebrowser.__version__)
+        lines = ['The report has been sent successfully. Thanks!']
+        if new_version > cur_version:
+            lines.append("<b>Note:</b> The newest available version is v{}, "
+                         "but you're currently running v{} - please "
+                         "update!".format(newest, qutebrowser.__version__))
+        text = '<br/><br/>'.join(lines)
+        msgbox.information(self, "Report successfully sent!", text,
+                           on_finished=self.finish, plain_text=False)
+
+    @pyqtSlot(str)
+    def on_version_error(self, msg):
+        """Called when the version was not obtained from self._pypi_client.
+
+        Args:
+            msg: The error message to show.
+        """
+        lines = ['The report has been sent successfully. Thanks!']
+        lines.append("There was an error while getting the newest version: "
+                     "{}. Please check for a new version on "
+                     "<a href=https://www.qutebrowser.org/>qutebrowser.org</a> "
+                     "by yourself.".format(msg))
+        text = '<br/><br/>'.join(lines)
+        msgbox.information(self, "Report successfully sent!", text,
+                           on_finished=self.finish, plain_text=False)
+
     @pyqtSlot()
     def finish(self):
         """Save contact info and close the dialog."""
@@ -325,17 +382,17 @@ class ExceptionCrashDialog(_CrashDialog):
         _pages: A list of lists of the open pages (URLs as strings)
         _cmdhist: A list with the command history (as strings)
         _exc: An exception tuple (type, value, traceback)
-        _objects: A list of all QObjects as string.
+        _qobjects: A list of all QObjects as string.
     """
 
-    def __init__(self, debug, pages, cmdhist, exc, objects, parent=None):
+    def __init__(self, debug, pages, cmdhist, exc, qobjects, parent=None):
         self._chk_log = None
         self._chk_restore = None
         super().__init__(debug, parent)
         self._pages = pages
         self._cmdhist = cmdhist
         self._exc = exc
-        self._objects = objects
+        self._qobjects = qobjects
         self.setModal(True)
         self._set_crash_info()
 
@@ -344,9 +401,6 @@ class ExceptionCrashDialog(_CrashDialog):
         text = "<b>Argh! qutebrowser crashed unexpectedly.</b>"
         self._lbl.setText(text)
 
-    def _init_buttons(self):
-        super()._init_buttons()
-
     def _init_checkboxes(self):
         """Add checkboxes to the dialog."""
         super()._init_checkboxes()
@@ -356,7 +410,7 @@ class ExceptionCrashDialog(_CrashDialog):
         self._chk_log = QCheckBox("Include a debug log in the report",
                                   checked=True)
         try:
-            if config.get('general', 'private-browsing'):
+            if config.val.content.private_browsing:
                 self._chk_log.setChecked(False)
         except Exception:
             log.misc.exception("Error while checking private browsing mode")
@@ -385,35 +439,37 @@ class ExceptionCrashDialog(_CrashDialog):
                 ("Commandline args", ' '.join(sys.argv[1:])),
                 ("Open Pages", '\n\n'.join('\n'.join(e) for e in self._pages)),
                 ("Command history", '\n'.join(self._cmdhist)),
-                ("Objects", self._objects),
+                ("Objects", self._qobjects),
             ]
             try:
                 self._crash_info.append(
                     ("Debug log", log.ram_handler.dump_log()))
             except Exception:
-                self._crash_info.append(
-                    ("Debug log", traceback.format_exc()))
+                self._crash_info.append(("Debug log", traceback.format_exc()))
 
     @pyqtSlot()
     def finish(self):
         self._save_contact_info()
         if self._chk_restore.isChecked():
-            self.accept()
+            self.done(Result.restore)
         else:
-            self.reject()
+            self.done(Result.no_restore)
 
 
 class FatalCrashDialog(_CrashDialog):
 
-    """Dialog which gets shown when a fatal error occured.
+    """Dialog which gets shown when a fatal error occurred.
 
     Attributes:
         _log: The log text to display.
-        _type: The type of error which occured.
-        _func: The function (top of the stack) in which the error occured.
+        _type: The type of error which occurred.
+        _func: The function (top of the stack) in which the error occurred.
+        _chk_history: A checkbox for the user to decide if page history should
+                      be sent.
     """
 
     def __init__(self, debug, text, parent=None):
+        self._chk_history = None
         super().__init__(debug, parent)
         self._log = text
         self.setAttribute(Qt.WA_DeleteOnClose)
@@ -421,7 +477,8 @@ class FatalCrashDialog(_CrashDialog):
         self._type, self._func = parse_fatal_stacktrace(self._log)
 
     def _get_error_type(self):
-        if self._type == 'Segmentation fault':
+        if self._type in ['Segmentation fault',
+                          'Windows access violation']:
             return 'segv'
         else:
             return self._type
@@ -439,9 +496,47 @@ class FatalCrashDialog(_CrashDialog):
                 "stacktrace.asciidoc</a> to submit a stacktrace.<br/>")
         self._lbl.setText(text)
 
+    def _init_checkboxes(self):
+        """Add checkboxes to the dialog."""
+        super()._init_checkboxes()
+        self._chk_history = QCheckBox("Include a history of the last "
+                                      "accessed pages in the report.",
+                                      checked=True)
+        try:
+            if config.val.content.private_browsing:
+                self._chk_history.setChecked(False)
+        except Exception:
+            log.misc.exception("Error while checking private browsing mode")
+        self._chk_history.toggled.connect(self._set_crash_info)
+        self._vbox.addWidget(self._chk_history)
+
     def _gather_crash_info(self):
         self._crash_info.append(("Fault log", self._log))
         super()._gather_crash_info()
+        if self._chk_history.isChecked():
+            try:
+                history = objreg.get('web-history').get_recent()
+                self._crash_info.append(("History",
+                                         '\n'.join(str(e) for e in history)))
+            except Exception:
+                self._crash_info.append(("History", traceback.format_exc()))
+
+    @pyqtSlot()
+    def on_report_clicked(self):
+        """Prevent empty reports."""
+        if (not self._info.toPlainText().strip() and
+                not self._contact.toPlainText().strip() and
+                self._get_error_type() == 'segv' and
+                self._func == 'qt_mainloop'):
+            msgbox.msgbox(parent=self, title='Empty crash info',
+                          text="Empty reports for fatal crashes are useless "
+                          "and mean I'll spend time deleting reports I could "
+                          "spend on developing qutebrowser instead.\n\nPlease "
+                          "help making qutebrowser better by providing more "
+                          "information, or don't report this.",
+                          icon=QMessageBox.Critical)
+        else:
+            super().on_report_clicked()
 
 
 class ReportDialog(_CrashDialog):
@@ -451,15 +546,15 @@ class ReportDialog(_CrashDialog):
     Attributes:
         _pages: A list of the open pages (URLs as strings)
         _cmdhist: A list with the command history (as strings)
-        _objects: A list of all QObjects as string.
+        _qobjects: A list of all QObjects as string.
     """
 
-    def __init__(self, pages, cmdhist, objects, parent=None):
+    def __init__(self, pages, cmdhist, qobjects, parent=None):
         super().__init__(False, parent)
         self.setAttribute(Qt.WA_DeleteOnClose)
         self._pages = pages
         self._cmdhist = cmdhist
-        self._objects = objects
+        self._qobjects = qobjects
         self._set_crash_info()
 
     def _init_text(self):
@@ -480,7 +575,7 @@ class ReportDialog(_CrashDialog):
             ("Commandline args", ' '.join(sys.argv[1:])),
             ("Open Pages", '\n\n'.join('\n'.join(e) for e in self._pages)),
             ("Command history", '\n'.join(self._cmdhist)),
-            ("Objects", self._objects),
+            ("Objects", self._qobjects),
         ]
         try:
             self._crash_info.append(("Debug log", log.ram_handler.dump_log()))
@@ -514,3 +609,42 @@ class ReportErrorDialog(QDialog):
         btn.clicked.connect(self.close)
         hbox.addWidget(btn)
         vbox.addLayout(hbox)
+
+
+def dump_exception_info(exc, pages, cmdhist, qobjects):
+    """Dump exception info to stderr.
+
+    Args:
+        exc: An exception tuple (type, value, traceback)
+        pages: A list of lists of the open pages (URLs as strings)
+        cmdhist: A list with the command history (as strings)
+        qobjects: A list of all QObjects as string.
+    """
+    print(file=sys.stderr)
+    print("\n\n===== Handling exception with --no-err-windows... =====\n\n",
+          file=sys.stderr)
+    print("\n---- Exceptions ----", file=sys.stderr)
+    print(''.join(traceback.format_exception(*exc)), file=sys.stderr)
+    print("\n---- Version info ----", file=sys.stderr)
+    try:
+        print(version.version(), file=sys.stderr)
+    except Exception:
+        traceback.print_exc()
+    print("\n---- Config ----", file=sys.stderr)
+    try:
+        print(config.instance.dump_userconfig(), file=sys.stderr)
+    except Exception:
+        traceback.print_exc()
+    print("\n---- Commandline args ----", file=sys.stderr)
+    print(' '.join(sys.argv[1:]), file=sys.stderr)
+    print("\n---- Open pages ----", file=sys.stderr)
+    print('\n\n'.join('\n'.join(e) for e in pages), file=sys.stderr)
+    print("\n---- Command history ----", file=sys.stderr)
+    print('\n'.join(cmdhist), file=sys.stderr)
+    print("\n---- Objects ----", file=sys.stderr)
+    print(qobjects, file=sys.stderr)
+    print("\n---- Environment ----", file=sys.stderr)
+    try:
+        print(_get_environment_vars(), file=sys.stderr)
+    except Exception:
+        traceback.print_exc()
