@@ -1,6 +1,6 @@
 # vim: ft=python fileencoding=utf-8 sts=4 sw=4 et:
 
-# Copyright 2014-2015 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
+# Copyright 2014-2017 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
 #
 # This file is part of qutebrowser.
 #
@@ -28,15 +28,17 @@ Module attributes:
 
 
 import io
-import os
-import sys
 import operator
-import distutils.version  # pylint: disable=no-name-in-module,import-error
-# https://bitbucket.org/logilab/pylint/issue/73/
 import contextlib
 
+import pkg_resources
 from PyQt5.QtCore import (qVersion, QEventLoop, QDataStream, QByteArray,
-                          QIODevice, QSaveFile)
+                          QIODevice, QSaveFile, QT_VERSION_STR,
+                          PYQT_VERSION_STR)
+try:
+    from PyQt5.QtWebKit import qWebKitVersion
+except ImportError:  # pragma: no cover
+    qWebKitVersion = None
 
 
 MAXVALS = {
@@ -50,17 +52,57 @@ MINVALS = {
 }
 
 
-def version_check(version, op=operator.ge):
+class QtOSError(OSError):
+
+    """An OSError triggered by a QFileDevice.
+
+    Attributes:
+        qt_errno: The error attribute of the given QFileDevice, if applicable.
+    """
+
+    def __init__(self, dev, msg=None):
+        if msg is None:
+            msg = dev.errorString()
+
+        super().__init__(msg)
+
+        try:
+            self.qt_errno = dev.error()
+        except AttributeError:
+            self.qt_errno = None
+
+
+def version_check(version, exact=False, compiled=True):
     """Check if the Qt runtime version is the version supplied or newer.
 
     Args:
         version: The version to check against.
-        op: The operator to use for the check.
+        exact: if given, check with == instead of >=
+        compiled: Set to False to not check the compiled version.
     """
-    # pylint: disable=no-member
-    # https://bitbucket.org/logilab/pylint/issue/73/
-    return op(distutils.version.StrictVersion(qVersion()),
-              distutils.version.StrictVersion(version))
+    # Catch code using the old API for this
+    assert exact not in [operator.gt, operator.lt, operator.ge, operator.le,
+                         operator.eq], exact
+    if compiled and exact:
+        raise ValueError("Can't use compiled=True with exact=True!")
+
+    parsed = pkg_resources.parse_version(version)
+    op = operator.eq if exact else operator.ge
+    result = op(pkg_resources.parse_version(qVersion()), parsed)
+    if compiled and result:
+        # qVersion() ==/>= parsed, now check if QT_VERSION_STR ==/>= parsed.
+        result = op(pkg_resources.parse_version(QT_VERSION_STR), parsed)
+    if compiled and result:
+        # FInally, check PYQT_VERSION_STR as well.
+        result = op(pkg_resources.parse_version(PYQT_VERSION_STR), parsed)
+    return result
+
+
+def is_new_qtwebkit():
+    """Check if the given version is a new QtWebKit."""
+    assert qWebKitVersion is not None
+    return (pkg_resources.parse_version(qWebKitVersion()) >
+            pkg_resources.parse_version('538.1'))
 
 
 def check_overflow(arg, ctype, fatal=True):
@@ -69,7 +111,7 @@ def check_overflow(arg, ctype, fatal=True):
     Args:
         arg: The argument to check
         ctype: The C/Qt type to check as a string.
-        fatal: Wether to raise exceptions (True) or truncate values (False)
+        fatal: Whether to raise exceptions (True) or truncate values (False)
 
     Return
         The truncated argument if fatal=False
@@ -91,49 +133,13 @@ def check_overflow(arg, ctype, fatal=True):
         return arg
 
 
-def get_args(namespace):
-    """Get the Qt QApplication arguments based on an argparse namespace.
-
-    Args:
-        namespace: The argparse namespace.
-
-    Return:
-        The argv list to be passed to Qt.
-    """
-    argv = [sys.argv[0]]
-    for argname, val in vars(namespace).items():
-        if not argname.startswith('qt_'):
-            continue
-        elif val is None:
-            # flag/argument not given
-            continue
-        elif val is True:
-            argv.append('-' + argname[3:])
-        else:
-            argv.append('-' + argname[3:])
-            argv.append(val[0])
-    return argv
-
-
-def check_print_compat():
-    """Check if printing should work in the given Qt version."""
-    # WORKAROUND (remove this when we bump the requirements to 5.3.0)
-    return not (os.name == 'nt' and version_check('5.3.0', operator.lt))
-
-
 def ensure_valid(obj):
     """Ensure a Qt object with an .isValid() method is valid."""
     if not obj.isValid():
         raise QtValueError(obj)
 
 
-def ensure_not_null(obj):
-    """Ensure a Qt object with an .isNull() method is not null."""
-    if obj.isNull():
-        raise QtValueError(obj)
-
-
-def _check_qdatastream(stream):
+def check_qdatastream(stream):
     """Check the status of a QDataStream and raise OSError if it's not ok."""
     status_to_str = {
         QDataStream.Ok: "The data stream is operating normally.",
@@ -151,41 +157,53 @@ def serialize(obj):
     """Serialize an object into a QByteArray."""
     data = QByteArray()
     stream = QDataStream(data, QIODevice.WriteOnly)
-    stream << obj  # pylint: disable=pointless-statement
-    _check_qdatastream(stream)
+    serialize_stream(stream, obj)
     return data
 
 
 def deserialize(data, obj):
     """Deserialize an object from a QByteArray."""
     stream = QDataStream(data, QIODevice.ReadOnly)
+    deserialize_stream(stream, obj)
+
+
+def serialize_stream(stream, obj):
+    """Serialize an object into a QDataStream."""
+    check_qdatastream(stream)
+    stream << obj  # pylint: disable=pointless-statement
+    check_qdatastream(stream)
+
+
+def deserialize_stream(stream, obj):
+    """Deserialize a QDataStream into an object."""
+    check_qdatastream(stream)
     stream >> obj  # pylint: disable=pointless-statement
-    _check_qdatastream(stream)
+    check_qdatastream(stream)
 
 
 @contextlib.contextmanager
 def savefile_open(filename, binary=False, encoding='utf-8'):
     """Context manager to easily use a QSaveFile."""
     f = QSaveFile(filename)
-    new_f = None
+    cancelled = False
     try:
-        ok = f.open(QIODevice.WriteOnly)
-        if not ok:  # pylint: disable=used-before-assignment
-            raise OSError(f.errorString())
+        open_ok = f.open(QIODevice.WriteOnly)
+        if not open_ok:
+            raise QtOSError(f)
         if binary:
             new_f = PyQIODevice(f)
         else:
             new_f = io.TextIOWrapper(PyQIODevice(f), encoding=encoding)
         yield new_f
+        new_f.flush()
     except:
         f.cancelWriting()
+        cancelled = True
         raise
     finally:
-        if new_f is not None:
-            new_f.flush()
-        ok = f.commit()
-        if not ok:
-            raise OSError(f.errorString())
+        commit_ok = f.commit()
+        if not commit_ok and not cancelled:
+            raise QtOSError(f, msg="Commit failed!")
 
 
 class PyQIODevice(io.BufferedIOBase):
@@ -193,27 +211,57 @@ class PyQIODevice(io.BufferedIOBase):
     """Wrapper for a QIODevice which provides a python interface.
 
     Attributes:
-        _dev: The underlying QIODevice.
+        dev: The underlying QIODevice.
     """
 
-    # pylint: disable=missing-docstring
-
     def __init__(self, dev):
-        self._dev = dev
+        super().__init__()
+        self.dev = dev
 
     def __len__(self):
-        return self._dev.size()
+        return self.dev.size()
 
     def _check_open(self):
-        """Check if the device is open, raise OSError if not."""
-        if not self._dev.isOpen():
-            raise OSError("IO operation on closed device!")
+        """Check if the device is open, raise ValueError if not."""
+        if not self.dev.isOpen():
+            raise ValueError("IO operation on closed device!")
 
     def _check_random(self):
         """Check if the device supports random access, raise OSError if not."""
         if not self.seekable():
             raise OSError("Random access not allowed!")
 
+    def _check_readable(self):
+        """Check if the device is readable, raise OSError if not."""
+        if not self.dev.isReadable():
+            raise OSError("Trying to read unreadable file!")
+
+    def _check_writable(self):
+        """Check if the device is writable, raise OSError if not."""
+        if not self.writable():
+            raise OSError("Trying to write to unwritable file!")
+
+    def open(self, mode):
+        """Open the underlying device and ensure opening succeeded.
+
+        Raises OSError if opening failed.
+
+        Args:
+            mode: QIODevice::OpenMode flags.
+
+        Return:
+            A contextlib.closing() object so this can be used as
+            contextmanager.
+        """
+        ok = self.dev.open(mode)
+        if not ok:
+            raise QtOSError(self.dev)
+        return contextlib.closing(self)
+
+    def close(self):
+        """Close the underlying device."""
+        self.dev.close()
+
     def fileno(self):
         raise io.UnsupportedOperation
 
@@ -221,73 +269,87 @@ class PyQIODevice(io.BufferedIOBase):
         self._check_open()
         self._check_random()
         if whence == io.SEEK_SET:
-            ok = self._dev.seek(offset)
+            ok = self.dev.seek(offset)
         elif whence == io.SEEK_CUR:
-            ok = self._dev.seek(self.tell() + offset)
+            ok = self.dev.seek(self.tell() + offset)
         elif whence == io.SEEK_END:
-            ok = self._dev.seek(len(self) + offset)
+            ok = self.dev.seek(len(self) + offset)
         else:
             raise io.UnsupportedOperation("whence = {} is not "
                                           "supported!".format(whence))
         if not ok:
-            raise OSError(self._dev.errorString())
+            raise QtOSError(self.dev, msg="seek failed!")
 
     def truncate(self, size=None):  # pylint: disable=unused-argument
         raise io.UnsupportedOperation
 
-    def close(self):
-        self._dev.close()
-
     @property
     def closed(self):
-        return not self._dev.isOpen()
+        return not self.dev.isOpen()
 
     def flush(self):
         self._check_open()
-        self._dev.waitForBytesWritten(-1)
+        self.dev.waitForBytesWritten(-1)
 
     def isatty(self):
         self._check_open()
         return False
 
     def readable(self):
-        return self._dev.isReadable()
+        return self.dev.isReadable()
 
     def readline(self, size=-1):
         self._check_open()
-        if size == -1:
-            size = 0
-        return self._dev.readLine(size)
+        self._check_readable()
+
+        if size < 0:
+            qt_size = 0  # no maximum size
+        elif size == 0:
+            return QByteArray()
+        else:
+            qt_size = size + 1  # Qt also counts the NUL byte
+
+        if self.dev.canReadLine():
+            buf = self.dev.readLine(qt_size)
+        else:
+            if size < 0:
+                buf = self.dev.readAll()
+            else:
+                buf = self.dev.read(size)
+
+        if buf is None:
+            raise QtOSError(self.dev)
+        return buf
 
     def seekable(self):
-        return not self._dev.isSequential()
+        return not self.dev.isSequential()
 
     def tell(self):
         self._check_open()
         self._check_random()
-        return self._dev.pos()
+        return self.dev.pos()
 
     def writable(self):
-        return self._dev.isWritable()
-
-    def readinto(self, b):
-        self._check_open()
-        return self._dev.read(b, len(b))
+        return self.dev.isWritable()
 
     def write(self, b):
         self._check_open()
-        num = self._dev.write(b)
+        self._check_writable()
+        num = self.dev.write(b)
         if num == -1 or num < len(b):
-            raise OSError(self._dev.errorString())
+            raise QtOSError(self.dev)
         return num
 
-    def read(self, size):
+    def read(self, size=-1):
         self._check_open()
-        buf = bytes()
-        num = self._dev.read(buf, size)
-        if num == -1:
-            raise OSError(self._dev.errorString())
-        return num
+        self._check_readable()
+        if size < 0:
+            buf = self.dev.readAll()
+        else:
+            buf = self.dev.read(size)
+        if buf is None:
+            raise QtOSError(self.dev)
+        return buf
 
 
 class QtValueError(ValueError):
