@@ -1,6 +1,6 @@
 # vim: ft=python fileencoding=utf-8 sts=4 sw=4 et:
 
-# Copyright 2014-2015 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
+# Copyright 2014-2017 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
 #
 # This file is part of qutebrowser.
 #
@@ -17,23 +17,71 @@
 # You should have received a copy of the GNU General Public License
 # along with qutebrowser.  If not, see <http://www.gnu.org/licenses/>.
 
-"""Other utilities which don't fit anywhere else. """
+"""Other utilities which don't fit anywhere else."""
 
+import os
+import os.path
 import io
+import re
 import sys
 import enum
-import inspect
-import os.path
+import json
 import collections
+import datetime
+import traceback
 import functools
 import contextlib
+import socket
+import shlex
 
-from PyQt5.QtCore import Qt
-from PyQt5.QtGui import QKeySequence, QColor
+import attr
+from PyQt5.QtCore import Qt, QUrl
+from PyQt5.QtGui import QKeySequence, QColor, QClipboard, QDesktopServices
+from PyQt5.QtWidgets import QApplication
 import pkg_resources
+import yaml
+try:
+    from yaml import CSafeLoader as YamlLoader, CSafeDumper as YamlDumper
+    YAML_C_EXT = True
+except ImportError:  # pragma: no cover
+    from yaml import SafeLoader as YamlLoader, SafeDumper as YamlDumper
+    YAML_C_EXT = False
 
 import qutebrowser
-from qutebrowser.utils import qtutils, log
+from qutebrowser.utils import qtutils, log, debug
+
+
+fake_clipboard = None
+log_clipboard = False
+
+is_mac = sys.platform.startswith('darwin')
+is_linux = sys.platform.startswith('linux')
+is_windows = sys.platform.startswith('win')
+is_posix = os.name == 'posix'
+
+
+class Unreachable(Exception):
+
+    """Raised when there was unreachable code."""
+
+
+class ClipboardError(Exception):
+
+    """Raised if the clipboard contents are unavailable for some reason."""
+
+
+class SelectionUnsupportedError(ClipboardError):
+
+    """Raised if [gs]et_clipboard is used and selection=True is unsupported."""
+
+    def __init__(self):
+        super().__init__("Primary selection is not supported on this "
+                         "platform!")
+
+
+class ClipboardEmptyError(ClipboardError):
+
+    """Raised if get_clipboard is used and the clipboard is empty."""
 
 
 def elide(text, length):
@@ -46,19 +94,49 @@ def elide(text, length):
         return text[:length - 1] + '\u2026'
 
 
+def elide_filename(filename, length):
+    """Elide a filename to the given length.
+
+    The difference to the elide() is that the text is removed from
+    the middle instead of from the end. This preserves file name extensions.
+    Additionally, standard ASCII dots are used ("...") instead of the unicode
+    "…" (U+2026) so it works regardless of the filesystem encoding.
+
+    This function does not handle path separators.
+
+    Args:
+        filename: The filename to elide.
+        length: The maximum length of the filename, must be at least 3.
+
+    Return:
+        The elided filename.
+    """
+    elidestr = '...'
+    if length < len(elidestr):
+        raise ValueError('length must be greater or equal to 3')
+    if len(filename) <= length:
+        return filename
+    # Account for '...'
+    length -= len(elidestr)
+    left = length // 2
+    right = length - left
+    if right == 0:
+        return filename[:left] + elidestr
+    else:
+        return filename[:left] + elidestr + filename[-right:]
+
+
 def compact_text(text, elidelength=None):
     """Remove leading whitespace and newlines from a text and maybe elide it.
 
-    FIXME: Add tests.
-
     Args:
         text: The text to compact.
         elidelength: To how many chars to elide.
     """
-    out = []
+    lines = []
     for line in text.splitlines():
-        out.append(line.strip())
-    out = ''.join(out)
+        lines.append(line.strip())
+    out = ''.join(lines)
     if elidelength is not None:
         out = elide(out, elidelength)
     return out
@@ -76,7 +154,8 @@ def read_file(filename, binary=False):
         The file contents as string.
     """
     if hasattr(sys, 'frozen'):
-        # cx_Freeze doesn't support pkg_resources :(
+        # PyInstaller doesn't support pkg_resources :(
+        # https://github.com/pyinstaller/pyinstaller/wiki/FAQ#misc
         fn = os.path.join(os.path.dirname(sys.executable), filename)
         if binary:
             with open(fn, 'rb') as f:
@@ -91,35 +170,18 @@ def read_file(filename, binary=False):
         return data
 
 
-def actute_warning():
-    """Display a warning about the dead_actute issue if needed."""
-    # WORKAROUND (remove this when we bump the requirements to 5.3.0)
-    # Non linux OS' aren't affected
-    if not sys.platform.startswith('linux'):
-        return
-    # If no compose file exists for some reason, we're not affected
-    if not os.path.exists('/usr/share/X11/locale/en_US.UTF-8/Compose'):
-        return
-    # Qt >= 5.3 doesn't seem to be affected
-    try:
-        if qtutils.version_check('5.3.0'):
-            return
-    except ValueError:
-        pass
-    try:
-        with open('/usr/share/X11/locale/en_US.UTF-8/Compose', 'r',
-                  encoding='utf-8') as f:
-            for line in f:
-                if '<dead_actute>' in line:
-                    if sys.stdout is not None:
-                        sys.stdout.flush()
-                    print("Note: If you got a 'dead_actute' warning above, "
-                          "that is not a bug in qutebrowser! See "
-                          "https://bugs.freedesktop.org/show_bug.cgi?id=69476 "
-                          "for details.")
-                    break
-    except OSError:
-        log.misc.exception("Failed to read Compose file")
+def resource_filename(filename):
+    """Get the absolute filename of a file contained with qutebrowser.
+
+    Args:
+        filename: The filename.
+
+    Return:
+        The absolute filename.
+    """
+    if hasattr(sys, 'frozen'):
+        return os.path.join(os.path.dirname(sys.executable), filename)
+    return pkg_resources.resource_filename(qutebrowser.__name__, filename)
 
 
 def _get_color_percentage(a_c1, a_c2, a_c3, b_c1, b_c2, b_c3, percent):
@@ -150,14 +212,22 @@ def interpolate_color(start, end, percent, colorspace=QColor.Rgb):
         start: The start color.
         end: The end color.
         percent: Which value to get (0 - 100)
-        colorspace: The desired interpolation colorsystem,
+        colorspace: The desired interpolation color system,
                     QColor::{Rgb,Hsv,Hsl} (from QColor::Spec enum)
+                    If None, start is used except when percent is 100.
 
     Return:
         The interpolated QColor, with the same spec as the given start color.
     """
     qtutils.ensure_valid(start)
     qtutils.ensure_valid(end)
+
+    if colorspace is None:
+        if percent == 100:
+            return QColor(*end.getRgb())
+        else:
+            return QColor(*start.getRgb())
+
     out = QColor()
     if colorspace == QColor.Rgb:
         a_c1, a_c2, a_c3, _alpha = start.getRgb()
@@ -226,7 +296,7 @@ def key_to_string(key):
     """
     special_names_str = {
         # Some keys handled in a weird way by QKeySequence::toString.
-        # See https://bugreports.qt-project.org/browse/QTBUG-40030
+        # See https://bugreports.qt.io/browse/QTBUG-40030
         # Most are unlikely to be ever needed, but you never know ;)
         # For dead/combining keys, we return the corresponding non-combining
         # key, as that's easier to add to the config.
@@ -274,6 +344,18 @@ def key_to_string(key):
         'Key_TouchpadOn': 'Touchpad On',
         'Key_TouchpadToggle': 'Touchpad toggle',
         'Key_Yellow': 'Yellow',
+        'Key_Alt': 'Alt',
+        'Key_AltGr': 'AltGr',
+        'Key_Control': 'Control',
+        'Key_Direction_L': 'Direction L',
+        'Key_Direction_R': 'Direction R',
+        'Key_Hyper_L': 'Hyper L',
+        'Key_Hyper_R': 'Hyper R',
+        'Key_Meta': 'Meta',
+        'Key_Shift': 'Shift',
+        'Key_Super_L': 'Super L',
+        'Key_Super_R': 'Super R',
+        'Key_unknown': 'Unknown',
     }
     # We now build our real special_names dict from the string mapping above.
     # The reason we don't do this directly is that certain Qt versions don't
@@ -310,21 +392,27 @@ def keyevent_to_string(e):
         A name of the key (combination) as a string or
         None if only modifiers are pressed..
     """
-    modmask2str = collections.OrderedDict([
-        (Qt.ControlModifier, 'Ctrl'),
-        (Qt.AltModifier, 'Alt'),
-        (Qt.MetaModifier, 'Meta'),
-        (Qt.ShiftModifier, 'Shift'),
-    ])
-    if sys.platform == 'darwin':
-        # FIXME verify this feels right on a real Mac as well.
-        # In my Virtualbox VM, the Ctrl key shows up as meta.
-        # https://github.com/The-Compiler/qutebrowser/issues/110
-        modmask2str[Qt.MetaModifier] = 'Ctrl'
+    if is_mac:
+        # Qt swaps Ctrl/Meta on macOS, so we switch it back here so the user
+        # can use it in the config as expected. See:
+        # https://github.com/qutebrowser/qutebrowser/issues/110
+        # http://doc.qt.io/qt-5.4/osx-issues.html#special-keys
+        modmask2str = collections.OrderedDict([
+            (Qt.MetaModifier, 'Ctrl'),
+            (Qt.AltModifier, 'Alt'),
+            (Qt.ControlModifier, 'Meta'),
+            (Qt.ShiftModifier, 'Shift'),
+        ])
+    else:
+        modmask2str = collections.OrderedDict([
+            (Qt.ControlModifier, 'Ctrl'),
+            (Qt.AltModifier, 'Alt'),
+            (Qt.MetaModifier, 'Meta'),
+            (Qt.ShiftModifier, 'Shift'),
+        ])
     modifiers = (Qt.Key_Control, Qt.Key_Alt, Qt.Key_Shift, Qt.Key_Meta,
-                 Qt.Key_AltGr, Qt.Key_Super_L, Qt.Key_Super_R,
-                 Qt.Key_Hyper_L, Qt.Key_Hyper_R, Qt.Key_Direction_L,
-                 Qt.Key_Direction_R)
+                 Qt.Key_AltGr, Qt.Key_Super_L, Qt.Key_Super_R, Qt.Key_Hyper_L,
+                 Qt.Key_Hyper_R, Qt.Key_Direction_L, Qt.Key_Direction_R)
     if e.key() in modifiers:
         # Only modifier pressed
         return None
@@ -334,7 +422,101 @@ def keyevent_to_string(e):
         if mod & mask and s not in parts:
             parts.append(s)
     parts.append(key_to_string(e.key()))
-    return '+'.join(parts)
+    return normalize_keystr('+'.join(parts))
+
+
+@attr.s(repr=False)
+class KeyInfo:
+
+    """Stores information about a key, like used in a QKeyEvent.
+
+    Attributes:
+        key: Qt::Key
+        modifiers: Qt::KeyboardModifiers
+        text: str
+    """
+
+    key = attr.ib()
+    modifiers = attr.ib()
+    text = attr.ib()
+
+    def __repr__(self):
+        if self.modifiers is None:
+            modifiers = None
+        else:
+            #modifiers = qflags_key(Qt, self.modifiers)
+            modifiers = hex(int(self.modifiers))
+        return get_repr(self, constructor=True,
+                        key=debug.qenum_key(Qt, self.key),
+                        modifiers=modifiers, text=self.text)
+
+
+class KeyParseError(Exception):
+
+    """Raised by _parse_single_key/parse_keystring on parse errors."""
+
+    def __init__(self, keystr, error):
+        super().__init__("Could not parse {!r}: {}".format(keystr, error))
+
+
+def is_special_key(keystr):
+    """True if keystr is a 'special' keystring (e.g. <ctrl-x> or <space>)."""
+    return keystr.startswith('<') and keystr.endswith('>')
+
+
+def _parse_single_key(keystr):
+    """Convert a single key string to a (Qt.Key, Qt.Modifiers, text) tuple."""
+    if is_special_key(keystr):
+        # Special key
+        keystr = keystr[1:-1]
+    elif len(keystr) == 1:
+        # vim-like key
+        pass
+    else:
+        raise KeyParseError(keystr, "Expecting either a single key or a "
+                            "<Ctrl-x> like keybinding.")
+
+    seq = QKeySequence(normalize_keystr(keystr), QKeySequence.PortableText)
+    if len(seq) != 1:
+        raise KeyParseError(keystr, "Got {} keys instead of 1.".format(
+            len(seq)))
+    result = seq[0]
+
+    if result == Qt.Key_unknown:
+        raise KeyParseError(keystr, "Got unknown key.")
+
+    modifier_mask = int(Qt.ShiftModifier | Qt.ControlModifier |
+                        Qt.AltModifier | Qt.MetaModifier | Qt.KeypadModifier |
+                        Qt.GroupSwitchModifier)
+    assert Qt.Key_unknown & ~modifier_mask == Qt.Key_unknown
+
+    modifiers = result & modifier_mask
+    key = result & ~modifier_mask
+
+    if len(keystr) == 1 and keystr.isupper():
+        modifiers |= Qt.ShiftModifier
+
+    assert key != 0, key
+    key = Qt.Key(key)
+    modifiers = Qt.KeyboardModifiers(modifiers)
+
+    # Let's hope this is accurate...
+    if len(keystr) == 1 and not modifiers:
+        text = keystr
+    elif len(keystr) == 1 and modifiers == Qt.ShiftModifier:
+        text = keystr.upper()
+    else:
+        text = ''
+
+    return KeyInfo(key, modifiers, text)
+
+
+def parse_keystring(keystr):
+    """Parse a keystring like <Ctrl-x> or xyz and return a KeyInfo list."""
+    if is_special_key(keystr):
+        return [_parse_single_key(keystr)]
+    else:
+        return [_parse_single_key(char) for char in keystr]
 
 
 def normalize_keystr(keystr):
@@ -346,17 +528,18 @@ def normalize_keystr(keystr):
     Return:
         The normalized keystring.
     """
+    keystr = keystr.lower()
     replacements = (
-        ('Control', 'Ctrl'),
-        ('Windows', 'Meta'),
-        ('Mod1', 'Alt'),
-        ('Mod4', 'Meta'),
+        ('control', 'ctrl'),
+        ('windows', 'meta'),
+        ('mod1', 'alt'),
+        ('mod4', 'meta'),
     )
     for (orig, repl) in replacements:
         keystr = keystr.replace(orig, repl)
-    for mod in ('Ctrl', 'Meta', 'Alt', 'Shift'):
+    for mod in ['ctrl', 'meta', 'alt', 'shift']:
         keystr = keystr.replace(mod + '-', mod + '+')
-    return keystr.lower()
+    return keystr
 
 
 class FakeIOStream(io.TextIOBase):
@@ -364,16 +547,9 @@ class FakeIOStream(io.TextIOBase):
     """A fake file-like stream which calls a function for write-calls."""
 
     def __init__(self, write_func):
+        super().__init__()
         self.write = write_func
 
-    def flush(self):
-        """This is only here to satisfy pylint."""
-        return super().flush()
-
-    def isatty(self):
-        """This is only here to satisfy pylint."""
-        return super().isatty()
-
 
 @contextlib.contextmanager
 def fake_io(write_func):
@@ -404,14 +580,16 @@ def disabled_excepthook():
     """Run code with the exception hook temporarily disabled."""
     old_excepthook = sys.excepthook
     sys.excepthook = sys.__excepthook__
-    yield
-    # If the code we did run did change sys.excepthook, we leave it
-    # unchanged. Otherwise, we reset it.
-    if sys.excepthook is sys.__excepthook__:
-        sys.excepthook = old_excepthook
+    try:
+        yield
+    finally:
+        # If the code we did run did change sys.excepthook, we leave it
+        # unchanged. Otherwise, we reset it.
+        if sys.excepthook is sys.__excepthook__:
+            sys.excepthook = old_excepthook
 
 
-class prevent_exceptions:  # pylint: disable=invalid-name
+class prevent_exceptions:  # noqa: N801,N806 pylint: disable=invalid-name
 
     """Decorator to ignore and log exceptions.
 
@@ -419,7 +597,7 @@ class prevent_exceptions:  # pylint: disable=invalid-name
     silently ignores them.
 
     We used to re-raise the exception with a single-shot QTimer in a similar
-    case, but that lead to a strange proble with a KeyError with some random
+    case, but that lead to a strange problem with a KeyError with some random
     jinja template stuff as content. For now, we only log it, so it doesn't
     pass 100% silently.
 
@@ -443,7 +621,7 @@ class prevent_exceptions:  # pylint: disable=invalid-name
         self._predicate = predicate
 
     def __call__(self, func):
-        """Gets called when a function should be decorated.
+        """Called when a function should be decorated.
 
         Args:
             func: The function to be decorated.
@@ -457,7 +635,8 @@ class prevent_exceptions:  # pylint: disable=invalid-name
         retval = self._retval
 
         @functools.wraps(func)
-        def wrapper(*args, **kwargs):  # pylint: disable=missing-docstring
+        def wrapper(*args, **kwargs):
+            """Call the original function."""
             try:
                 return func(*args, **kwargs)
             except BaseException:
@@ -486,7 +665,8 @@ def get_repr(obj, constructor=False, **attrs):
     """
     cls = qualname(obj.__class__)
     parts = []
-    for name, val in attrs.items():
+    items = sorted(attrs.items())
+    for name, val in items:
         parts.append('{}={!r}'.format(name, val))
     if constructor:
         return '{}({})'.format(cls, ', '.join(parts))
@@ -511,20 +691,18 @@ def qualname(obj):
     """
     if isinstance(obj, functools.partial):
         obj = obj.func
-    if hasattr(obj, '__qualname__'):
-        name = obj.__qualname__
-    elif hasattr(obj, '__name__'):
-        name = obj.__name__
-    else:
-        name = '<unknown>'
 
-    if inspect.isclass(obj) or inspect.isfunction(obj):
-        module = obj.__module__
-        return "{}.{}".format(module, name)
-    elif inspect.ismethod(obj):
-        return "{}.{}".format(obj.__module__, name)
+    if hasattr(obj, '__module__'):
+        prefix = '{}.'.format(obj.__module__)
     else:
-        return name
+        prefix = ''
+
+    if hasattr(obj, '__qualname__'):
+        return '{}{}'.format(prefix, obj.__qualname__)
+    elif hasattr(obj, '__name__'):
+        return '{}{}'.format(prefix, obj.__name__)
+    else:
+        return repr(obj)
 
 
 def raises(exc, func, *args):
@@ -552,3 +730,188 @@ def force_encoding(text, encoding):
     This replaces all chars not encodable with question marks.
     """
     return text.encode(encoding, errors='replace').decode(encoding)
+
+
+def sanitize_filename(name, replacement='_'):
+    """Replace invalid filename characters.
+
+    Note: This should be used for the basename, as it also removes the path
+    separator.
+
+    Args:
+        name: The filename.
+        replacement: The replacement character (or None).
+    """
+    if replacement is None:
+        replacement = ''
+    # Bad characters taken from Windows, there are even fewer on Linux
+    # See also
+    # https://en.wikipedia.org/wiki/Filename#Reserved_characters_and_words
+    bad_chars = '\\/:*?"<>|'
+    for bad_char in bad_chars:
+        name = name.replace(bad_char, replacement)
+    return name
+
+
+def set_clipboard(data, selection=False):
+    """Set the clipboard to some given data."""
+    if selection and not supports_selection():
+        raise SelectionUnsupportedError
+    if log_clipboard:
+        what = 'primary selection' if selection else 'clipboard'
+        log.misc.debug("Setting fake {}: {}".format(what, json.dumps(data)))
+    else:
+        mode = QClipboard.Selection if selection else QClipboard.Clipboard
+        QApplication.clipboard().setText(data, mode=mode)
+
+
+def get_clipboard(selection=False, fallback=False):
+    """Get data from the clipboard.
+
+    Args:
+        selection: Use the primary selection.
+        fallback: Fall back to the clipboard if primary selection is
+                  unavailable.
+    """
+    global fake_clipboard
+    if fallback and not selection:
+        raise ValueError("fallback given without selection!")
+
+    if selection and not supports_selection():
+        if fallback:
+            selection = False
+        else:
+            raise SelectionUnsupportedError
+
+    if fake_clipboard is not None:
+        data = fake_clipboard
+        fake_clipboard = None
+    else:
+        mode = QClipboard.Selection if selection else QClipboard.Clipboard
+        data = QApplication.clipboard().text(mode=mode)
+
+    target = "Primary selection" if selection else "Clipboard"
+    if not data.strip():
+        raise ClipboardEmptyError("{} is empty.".format(target))
+    log.misc.debug("{} contained: {!r}".format(target, data))
+
+    return data
+
+
+def supports_selection():
+    """Check if the OS supports primary selection."""
+    return QApplication.clipboard().supportsSelection()
+
+
+def random_port():
+    """Get a random free port."""
+    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+    sock.bind(('localhost', 0))
+    port = sock.getsockname()[1]
+    sock.close()
+    return port
+
+
+def open_file(filename, cmdline=None):
+    """Open the given file.
+
+    If cmdline is not given, downloads.open_dispatcher is used.
+    If open_dispatcher is unset, the system's default application is used.
+
+    Args:
+        filename: The filename to open.
+        cmdline: The command to use as string. A `{}` is expanded to the
+                 filename. None means to use the system's default application
+                 or `downloads.open_dispatcher` if set. If no `{}` is found,
+                 the filename is appended to the cmdline.
+    """
+    # Import late to avoid circular imports:
+    # - usertypes -> utils -> guiprocess -> message -> usertypes
+    # - usertypes -> utils -> config -> configdata -> configtypes ->
+    #   cmdutils -> command -> message -> usertypes
+    from qutebrowser.config import config
+    from qutebrowser.misc import guiprocess
+
+    # the default program to open downloads with - will be empty string
+    # if we want to use the default
+    override = config.val.downloads.open_dispatcher
+
+    # precedence order: cmdline > downloads.open_dispatcher > openUrl
+
+    if cmdline is None and not override:
+        log.misc.debug("Opening {} with the system application"
+                       .format(filename))
+        url = QUrl.fromLocalFile(filename)
+        QDesktopServices.openUrl(url)
+        return
+
+    if cmdline is None and override:
+        cmdline = override
+
+    cmd, *args = shlex.split(cmdline)
+    args = [arg.replace('{}', filename) for arg in args]
+    if '{}' not in cmdline:
+        args.append(filename)
+    log.misc.debug("Opening {} with {}"
+                   .format(filename, [cmd] + args))
+    proc = guiprocess.GUIProcess(what='open-file')
+    proc.start_detached(cmd, args)
+
+
+def unused(_arg):
+    """Function which does nothing to avoid pylint complaining."""
+    pass
+
+
+def expand_windows_drive(path):
+    r"""Expand a drive-path like E: into E:\.
+
+    Does nothing for other paths.
+
+    Args:
+        path: The path to expand.
+    """
+    # Usually, "E:" on Windows refers to the current working directory on drive
+    # E:\. The correct way to specifify drive E: is "E:\", but most users
+    # probably don't use the "multiple working directories" feature and expect
+    # "E:" and "E:\" to be equal.
+    if re.fullmatch(r'[A-Z]:', path, re.IGNORECASE):
+        return path + "\\"
+    else:
+        return path
+
+
+def yaml_load(f):
+    """Wrapper over yaml.load using the C loader if possible."""
+    start = datetime.datetime.now()
+    data = yaml.load(f, Loader=YamlLoader)
+    end = datetime.datetime.now()
+
+    delta = (end - start).total_seconds()
+    deadline = 5 if 'CI' in os.environ else 2
+    if delta > deadline:  # pragma: no cover
+        log.misc.warning(
+            "YAML load took unusually long, please report this at "
+            "https://github.com/qutebrowser/qutebrowser/issues/2777\n"
+            "duration: {}s\n"
+            "PyYAML version: {}\n"
+            "C extension: {}\n"
+            "Stack:\n\n"
+            "{}".format(
+                delta, yaml.__version__, YAML_C_EXT,
+                ''.join(traceback.format_stack())))
+
+    return data
+
+
+def yaml_dump(data, f=None):
+    """Wrapper over yaml.dump using the C dumper if possible.
+
+    Also returns a str instead of bytes.
+    """
+    yaml_data = yaml.dump(data, f, Dumper=YamlDumper, default_flow_style=False,
+                          encoding='utf-8', allow_unicode=True)
+    if yaml_data is None:
+        return None
+    else:
+        return yaml_data.decode('utf-8')
